/*
 * Roman Kollar, 396030
 * Michal Gajdar, 373872
 * DER certificate format parser */

/* Functions:
 *  Most of the code is composed from checkX functions, which
 *  parse the X tag in ASN.1. They either return true/false(or 1/0) or
 *  call the fail() function, which terminates the program.
 *  They take start of the buffer and position in it.
 *  Thus, documenting each function separately is pointless.
 *
 * Limitations:
 *  Program parses only a subset of the grammar. Eg. it does not print 
 *  algorithm names and it does not parse public keys. List of supported
 *  parts is marked in the der.png file.
 *
 *  Compilation:
 *   gcc -std=c99 -Wall -Wextra -lm DERParser.c -o der
 *
 *  Usage:
 *   ./der <input file>
 */


#include "stdio.h"
#include "stdlib.h"
#include "unistd.h"
#include "string.h"
#include "stdlib.h"
#include "math.h"
#include "stdint.h"
#include "stdbool.h"

#define BYTE 256

#define ZERO 0x00
#define INTEGER 0x02
#define BIT_STRING 0x03
#define NULLTYPE 0x05
#define OBJECT_ID 0x06
#define UTF8_STRING 0x08
#define PRINTABLE_STRING 0x13
#define T61_STRING 0x14
#define UTC_TIME 0x17
#define SEQUENCE 0x30
#define SET 0x31

static long fsize;

void printCurrentPosition(long *position);
void fail(void *);
void printHexArray(unsigned char* hexArray, int length, int depth);
void printString(unsigned char* string, int length);
int checkSequence(unsigned char *buf, long *position, int *length, int *number_of_bytes);
int intToString(uint64_t number, char** string);
int CheckTBSCertificate(unsigned char* buf, long *position);
int checkVersion(unsigned char* buf, long *position);
int checkInteger(unsigned char* buf, long *position);
int checkSerialNumber(unsigned char* buf, long *position, unsigned char** serialNumber);
int checkAlgorithmIdentifier(unsigned char* buf, long *position);
int checkObjectIdentifier(unsigned char* buf, long *position, char** objID);
int checkParameters(unsigned char* buf, long *position);
int checkName(unsigned char* buf, long *position);
int checkSet(unsigned char* buf, long *position, int *length, int* number_of_bytes);
int checkPrintableString(unsigned char* buf, long *position, unsigned char** string);
int checkT61String(unsigned char* buf, long *position, unsigned char** string);
int checkValidity(unsigned char* buf, long *position);
int checkUTCTime(unsigned char* buf, long *position, unsigned char** timeBuffer);
int checkSubject(unsigned char* buf, long *position);
bool checkSubjectPublicKeyInfo(unsigned char* buf, long *position);
bool checkUniqueIdentifier(unsigned char* buf, long *position);
bool checkExtensions(unsigned char* buf, long *position);
bool checkExtension(unsigned char* buf, long *position);
bool checkBitString(unsigned char *buf, long *position, int *length, int *number_of_bytes);

/* move the position if we are not at the end of the file
 * fail otherwise */
#define BUMP do { if ((*position) < fsize - 1) (*position)++; else fail(buf);} while(0)

void printOffset(int depth)
{
    for (int i = 0; i < depth; i++)
    {
        putchar('\t');
    }
}

static inline bool isContextSpecific(unsigned char x)
{
    return (x & 0xc0) == 0x80;
}


void printCurrentPosition(long *position)
{
	printf("Current position: %ld\n", *position);
}

uint64_t from_seq(unsigned char *in)
{
	uint64_t r = 0;

	do {
		r = (r << 7) | (uint64_t)(*in & 127);
	} while (*in++ & 128);

	return r;
}

/* check SEQUENCE tag */
int checkSequence(unsigned char *buf, long *position, int *length, int *number_of_bytes) {
	/*
	 * return 1 on success, zero otherwise
	 */
	if (*(buf + (*position)) != SEQUENCE)
	{
		return 0;
	}

    BUMP;
	int bytes = 0;

	if (*(buf + (*position)) < 0x80) {
		bytes = *(buf + (*position));
		BUMP;
	}
	else {
		*number_of_bytes = *(buf + (*position)) - 0x80;
		BUMP;
		for (int i = 0; i < *number_of_bytes; i++)
		{
			bytes += (int) *(buf + (*position)) * pow(BYTE, *number_of_bytes - 1 - i);
			BUMP;
		}
	}

	*length = bytes;
	return 1;
}

/* check BIT STRING tag */
bool checkBitString(unsigned char *buf, long *position, int *length, int *number_of_bytes) {
	if (*(buf + (*position)) != BIT_STRING)
	{
		return false;
	}

	BUMP;
	int bytes = 0;

	if (*(buf + (*position)) < 0x80) {
		bytes = *(buf + (*position));
        BUMP;
	}
	else {
		*number_of_bytes = *(buf + (*position)) - 0x80;
        BUMP;
		for (int i = 0; i < *number_of_bytes; i++)
		{
			bytes += (int) *(buf + (*position)) * pow(BYTE, *number_of_bytes - 1 - i);
            BUMP;
		}
	}

	*length = bytes;
	return true;
}


void fail(void *addr)
{
	free(addr);
	fprintf(stderr, "It is not proper DER file.\n");
	exit(EXIT_FAILURE);
}

/* print hex data with formatting offset */
void printHexArray(unsigned char* hexArray, int length, int depth)
{
	for (int i = 0; i < length; i++)
	{
        if (i % 20 == 0)
        {
            if (i != 0)
                putchar('\n');
            for (int j = 0; j < depth; j++)
                putchar('\t');
        }
		printf("%02x", hexArray[i]);
        if (i < length - 1)
            putchar(':');
	}
}

void printDecArray(unsigned char* hexArray, int length)
{
	for (int i = 0; i < length; i++)
	{
		printf("%d", hexArray[i]);
    }
}

void printString(unsigned char* string, int length)
{
	for (int i = 0; i < length; i++)
	{
		printf("%c", string[i]);
	}
}

/* print UTCTime tag format */
void printTime(unsigned char* string, int length)
{
    printf("20%c%c-", string[0], string[1]);
    printf("%c%c-", string[2], string[3]);
    printf("%c%c ", string[4], string[5]);
	for (int i = 6; i < length; i++)
	{
		printf("%c", string[i]);
	}
}

int checkTBSCertificate(unsigned char* buf, long *position)
{
//	printf("in start CheckTBSCertificate position: %d, value: %02x\n", *position, *(buf + *position));
	int length, number_of_bytes = 0;
	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		fail(buf);
	}

	int version = checkVersion(buf, position);
	printf("\tVersion: %d\n", version);

	unsigned char* serialNumber = NULL;
	int lengthOfSerialNumber = checkSerialNumber(buf, position, &serialNumber);
	printf("\tSerial Number: ");
	printDecArray(serialNumber, lengthOfSerialNumber);
	putchar('\n');
	free(serialNumber);
	
	printf("\tSignature:\n");
	checkAlgorithmIdentifier(buf, position);

	printf("\tIssuer:\n");
	if (!checkName(buf, position))
	{
		fail(buf);
	}

	printf("\tValidity:\n");
	if (!checkValidity(buf, position))
	{
		fail(buf);
	}

	printf("\tSubject:\n");
	if (!checkSubject(buf, position))
    {
        fail(buf);
    }

	printf("\tSubject public key info:\n");
    checkSubjectPublicKeyInfo(buf, position);

    if (isContextSpecific(*(buf + (*position))) && ((*(buf + (*position)) & 0xf) == 1))
    {
        printf("\tIssuer unique ID:\n");
        *(buf + (*position)) = BIT_STRING;
        checkUniqueIdentifier(buf, position);
    }
    if (isContextSpecific(*(buf + (*position))) && ((*(buf + (*position)) & 0xf) == 2))
    {
        printf("\tSubject unique ID:\n");
        *(buf + (*position)) = BIT_STRING;
        checkUniqueIdentifier(buf, position);
    }
    if (isContextSpecific(*(buf + (*position))) && ((*(buf + (*position)) & 0xf) == 3))
    {
        printf("\tX509v3 extensions:\n");
        *(buf + (*position)) = SEQUENCE;
        if (!checkSequence(buf, position, &length, &number_of_bytes))
            fail(buf);
        checkExtensions(buf, position);
    }

//	printf("in end CheckTBSCertificate\n");
	return 0;
}

int checkVersion(unsigned char* buf, long *position)
{
	int version = 0;	// DEFAULT version is v1 = 0
	// A0 03 Context Specific 0 with length 3
//	printf("in CheckVersion position: %d, value: %02x\n", *position, *(buf + *position));
	if ((*(buf + *position) == 0xA0) && (*(buf + (*position+1)) == 0x03))
	{
		*position += 2;
		version = checkInteger(buf, position);
	}
	return version;
}

int checkInteger(unsigned char* buf, long *position)
{
//	printf("in CheckInteger position: %d, value: %02x\n", *position, *(buf+*position));
	if (*(buf + *position) != INTEGER)
	{
		// exiting program
		fail(buf);
	}
	BUMP;

	int integer = 0;
	int number_of_bytes = *(buf + *position);

	BUMP;
	for (int i = 0; i < number_of_bytes; i++)
	{
		integer += (int)*(buf + *position) * pow(BYTE, number_of_bytes - 1 - i);
		BUMP;
	}

	return integer;
}

int checkSerialNumber(unsigned char* buf, long *position, unsigned char** serialNumber)
{
	// It is type INTEGER, very long in hexadecimal
//	printf("in CheckSerialNumber position: %d, value: %02x\n", *position, *(buf + *position));

	unsigned char* tmpSerialNumber = NULL;
	if (*(buf + (*position)) != INTEGER)
	{
		fail(buf);
	}
    BUMP;

	int number_of_bytes = *(buf + (*position));
    BUMP;

	//*serialNumber = (unsigned char*)malloc(number_of_bytes * sizeof(unsigned char));
	tmpSerialNumber = (unsigned char*)malloc(number_of_bytes * sizeof(unsigned char));
	for (int i = 0; i < number_of_bytes; i++)
	{
		//(*serialNumber)[i] = *(buf + (*position)++);
		tmpSerialNumber[i] = *(buf + (*position));
        BUMP;
	}
	*serialNumber = tmpSerialNumber;
	return number_of_bytes;
}

int checkAlgorithmIdentifier(unsigned char* buf, long *position)
{
	//AlgorithmIdentifier :: = SEQUENCE
	//printf("in start checkAlgorithmIdentifier position: %d, value: %02x\n", *position, *(buf + *position));
	
	int length, number_of_bytes = 0;
	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		fail(buf);
	}

	char* objID = NULL;
	
	//int lengthObjID = checkObjectIdentifier(buf, position, &objID);
	checkObjectIdentifier(buf, position, &objID);
	//printHexArray(objID, lengthObjID);
	
	printf("\t\tAlgorithm: %s\n", objID);
	free(objID);

	//parameters
	printf("\t\t\tParameters: ");
	if (checkParameters(buf, position))
	{
		printf("<not supported>\n");
	}
	else {
		printf("NULL\n");
	}


	return 1;
}

int checkObjectIdentifier(unsigned char* buf, long *position, char** objID)
{
//	printf("in start checkObjectIdentifier position: %d, value: %02x\n", *position, *(buf + *position));
	unsigned char* tmpObjID = NULL;
	if (*(buf + (*position)) != OBJECT_ID)
	{
		// exiting program
		fail(buf);
	}
    BUMP;

	int number_of_bytes = *(buf + (*position));
    BUMP;
	
	tmpObjID = (unsigned char*)malloc(number_of_bytes * sizeof(unsigned char));
	for (int i = 0; i < number_of_bytes; i++)
	{
		tmpObjID[i] = *(buf + (*position));
        BUMP;
	}

	char *bufObjID = NULL;

	int posBufIbjID = 0;
	int length = number_of_bytes * 4;			// better to do with cont. realloc
	bufObjID = (char*)realloc(bufObjID, length);
	
	int posTmpObjID = 0;
	int a = tmpObjID[posTmpObjID++];
	
	int first = a / 40;
	int second = a % 40;
	
	char* tmp = NULL;

	intToString(first, &tmp);
	memcpy((bufObjID + posBufIbjID), tmp, strlen(tmp));
	posBufIbjID = strlen(tmp);
	bufObjID[posBufIbjID++] = '.';
    free(tmp);
	
	intToString(second, &tmp);
	memcpy(bufObjID + posBufIbjID, tmp, strlen(tmp));
	posBufIbjID += strlen(tmp);
	bufObjID[posBufIbjID++] = '.';
    free(tmp);
	
	while (posTmpObjID < number_of_bytes)
	{
		if (tmpObjID[posTmpObjID] <= 127)
		{
			intToString(tmpObjID[posTmpObjID++], &tmp);
		}
		else
		{
			unsigned char* longVLQ = NULL;
			int lengthVLQ = 0;
			//longVLQ = (unsigned char*) realloc(longVLQ, sizeof(unsigned char) * lengthVLQ);
			while (tmpObjID[posTmpObjID] > 127)
			{
				if ((longVLQ = (unsigned char*)realloc(longVLQ, sizeof(unsigned char) * (lengthVLQ + 1))) == NULL) // longVLQ = (unsigned char*)realloc(longVLQ, sizeof(unsigned char) * (lengthVLQ + 1));
				{
					free(longVLQ);
					fail(buf);
				}
				longVLQ[lengthVLQ++] = tmpObjID[posTmpObjID++];
			}
			if ((longVLQ = (unsigned char*)realloc(longVLQ, sizeof(unsigned char) * (lengthVLQ + 1))) == NULL) 
			{
				free(longVLQ);
                fail(buf);
			}
			longVLQ[lengthVLQ] = tmpObjID[posTmpObjID++];

			intToString(from_seq(longVLQ), &tmp);
			free(longVLQ);
		}

		memcpy((bufObjID + posBufIbjID), tmp, strlen(tmp));
		posBufIbjID += strlen(tmp);
        free(tmp);
		bufObjID[posBufIbjID++] = '.';
	}
	
	bufObjID[--posBufIbjID] = '\0';

	free(tmpObjID);
	*objID = bufObjID;
	return number_of_bytes;
}

int intToString(uint64_t number, char** string)
{
	char* buffer = (char*)malloc(sizeof(char) * 10);
	int n = 0;
	n = sprintf(buffer, "%ld", number);

	*string = buffer;
	return n;
}

int checkParameters(unsigned char* buf, long *position)
{
	/* parameters field is optional by algorithm
	 * if there are any, it is type of SEQUENCE
	 * if not it is type of NULL with value 0x00
	 */

	int length = 0; 
	int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		if ((*(buf + (*position)++) == NULLTYPE) && (*(buf + (*position)++) == ZERO))
		{
			return 0;
		}
		else
		{
			fail(buf);
		}
	}
	(*position) += length;		// skipping SEQUENCE field

	return 1;
}

int checkName(unsigned char* buf, long *position)
{
	int length = 0;
	//int length_of_set = 0;
	int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes)) {
		return 0;
	}
    
    int total_length_of_sets = 0;

    while (total_length_of_sets < length)
    {
        int length_of_set = 0;
        int number_of_set_bytes = 0;
        if (checkSet(buf, position, &length_of_set, &number_of_set_bytes))
        {
            int seq_length = 0;
            int seq_number_of_bytes = 0;

            if (!checkSequence(buf, position, &seq_length, &seq_number_of_bytes))
            {
                return 0;
            }
            else
            {
                char* objID = NULL;
                checkObjectIdentifier(buf, position, &objID);

                // names
                printf("\t\t%s=", objID);
                free(objID);

                unsigned char* string = NULL;
                if (checkPrintableString(buf, position, &string))
                {
                    printf("%s ", string);
                    free(string);
                }
                else 
                {
                    int num;
                    if ((num = checkT61String(buf, position, &string)) >= 0)
                    {
                        printString(string, num);
                        free(string);
                        printf(" ");
                    }
                }
                putchar('\n');
            }

            //printf("length of set: %d \n", length_of_set);
            total_length_of_sets += length_of_set + number_of_set_bytes + 1;	// 1 stands for SET byte
            //printf("length: %d , total_lengt_of_sets: %d\n", length, total_length_of_sets);
        }
        else
        {
            return 0;
        }
    }

    //printf("end length: %d , total_lengt_of_sets: %d\n", length, total_length_of_sets);

	return 1;
}

int checkSet(unsigned char* buf, long *position, int *length, int* number_of_bytes)
{
	/*
	* return 1 on success, zero otherwise
	*/
	
	if (*(buf + (*position)) != SET)
	{
		return 0;
	}
	BUMP;

	int bytes = 0;
	//int number_of_bytes = 1;

	if (*(buf + (*position)) < 0x80) {
		bytes = *(buf + (*position));
        BUMP;
		*number_of_bytes = 1;
	}
	else {
		*number_of_bytes = *(buf + (*position)) - 0x80;
        BUMP;
		
		for (int i = 0; i < *number_of_bytes; i++)
		{
			bytes += (int)*(buf + (*position)) * pow(BYTE, *number_of_bytes - 1 - i);
            BUMP;
		}
	}

	*length = bytes;
	return 1;
}

int checkPrintableString(unsigned char* buf, long *position, unsigned char** string)
{
	unsigned char* tmpString = NULL;
	if (*(buf + (*position)) != PRINTABLE_STRING)
	{
		//fail(buf);
		return 0;
	}

	int number_of_bytes = 0;
	int bytes = 0;
	BUMP;

	if (*(buf + (*position)) < 0x80) {
		bytes = *(buf + (*position));
        BUMP;
		number_of_bytes++;
	}
	else {
		number_of_bytes = *(buf + (*position)) - 0x80;
		BUMP;
		for (int i = 0; i < number_of_bytes; i++)
		{
			bytes += (int)*(buf + (*position)) * pow(BYTE, number_of_bytes - 1 - i);
            BUMP;
		}
	}

	tmpString = (unsigned char*)malloc(bytes * sizeof(unsigned char) + 1);
	for (int i = 0; i < bytes; i++)
	{
		tmpString[i] = *(buf + (*position));
        BUMP;
	}

	tmpString[bytes] = '\0';
	*string = tmpString;

	return number_of_bytes;
}

int checkT61String(unsigned char* buf, long *position, unsigned char** string)
{
	unsigned char* tmpString = NULL;
	if (*(buf + (*position)) != T61_STRING)
	{
		//fail(buf);
		return -1;
	}
	//printf("In checkT61String\n");
	BUMP;
	int number_of_bytes = *(buf + (*position));
    BUMP;

	tmpString = (unsigned char*)malloc(number_of_bytes * sizeof(unsigned char));
	for (int i = 0; i < number_of_bytes; i++)
	{
		tmpString[i] = *(buf + (*position));
        BUMP;
	}
	*string = tmpString;
	return number_of_bytes;
}

int checkValidity(unsigned char* buf, long *position)
{
	int length = 0;
	int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		return 0;
	}

	unsigned char* timeBuffer = NULL;
	int lengthOfTimeBuffer = 0;
	
	/* NOT BEFORE TIME */
	lengthOfTimeBuffer = checkUTCTime(buf, position, &timeBuffer);
	if (!lengthOfTimeBuffer)
	{
		return 0;
	}

	printf("\t\tNot before: ");
	printTime(timeBuffer, lengthOfTimeBuffer);
	putchar('\n');
	free(timeBuffer);

	/* NOTAFTER TIME */
	lengthOfTimeBuffer = checkUTCTime(buf, position, &timeBuffer);
	if (!lengthOfTimeBuffer)
	{
		return 0;
	}
	printf("\t\tNot after:  ");
	printTime(timeBuffer, lengthOfTimeBuffer);
	putchar('\n');
	free(timeBuffer);

	return 1;
}

int checkUTCTime(unsigned char* buf, long *position, unsigned char** timeBuffer)
{
	unsigned char* tmpTimeBuff = NULL;
	if (*(buf + (*position)) != UTC_TIME)
	{
		fail(buf);
		//return 0;
	}
	//printf("In checkUTCTime\n");
	BUMP;
	int number_of_bytes = *(buf + (*position));
    BUMP;

	tmpTimeBuff = (unsigned char*)malloc(number_of_bytes * sizeof(unsigned char));
	for (int i = 0; i < number_of_bytes; i++)
	{
		tmpTimeBuff[i] = *(buf + (*position));
        BUMP;
	}
	*timeBuffer = tmpTimeBuff;
	return number_of_bytes;
}

int checkSubject(unsigned char* buf, long *position)
{
	int length = 0;
	int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		return 0;
	}

	int total_length_of_sets = 0;

	while (total_length_of_sets < length)
	{
		int length_of_set = 0;
		int number_of_set_bytes = 0;
		if (checkSet(buf, position, &length_of_set, &number_of_set_bytes))
		{
			int seq_length = 0;
			int seq_number_of_bytes = 0;

			if (!checkSequence(buf, position, &seq_length, &seq_number_of_bytes))
			{
				return 0;
			}
			else
			{
				char* objID = NULL;
				checkObjectIdentifier(buf, position, &objID);

				// names
				printf("\t\t%s=", objID);
				free(objID);

				unsigned char* string = NULL;
				if (checkPrintableString(buf, position, &string))
				{
					printf("%s \n", string);
                    free(string);
				}
				else 
				{
                    int num;
                    if ((num = checkT61String(buf, position, &string)) >= 0)
                    {
                        printString(string, num);
                        printf(" ");
                        free(string);
                    }
                    else 
                    {
                        printf("UTF8 string \n");
                        //printCurrentPosition(position);
                    }
                    putchar('\n');
				}

			}

			//printf("length of set: %d \n", length_of_set);
			total_length_of_sets += length_of_set + number_of_set_bytes + 1;	// 1 stands for SET byte
			//printf("length: %d , total_lengt_of_sets: %d\n", length, total_length_of_sets);
		}
		else
		{
			return 0;
		}
	}

	//printf("end length: %d , total_lengt_of_sets: %d\n", length, total_length_of_sets);
    return 1;
}

bool checkSubjectPublicKeyInfo(unsigned char* buf, long *position)
{
	int length = 0;
    int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
        fail(buf);
		return false;
	}
	
    if (!checkAlgorithmIdentifier(buf, position))
	{
        fail(buf);
        return false;
	}

    printf("\t\tSubject public key:\n");
	if (!checkBitString(buf, position, &length, &number_of_bytes))
	{
        fail(buf);
		return false;
	}

    printHexArray(buf + (*position), length, 3);
    printf("\n");
    *position += length;

    return true;
}

bool checkUniqueIdentifier(unsigned char* buf, long *position)
{
    int number_of_bytes = 0;

    if (*(buf + (*position)) != BIT_STRING) {
        printf("A\n");
        return false;
    }
   
    BUMP;
    number_of_bytes = *(buf + (*position));
    BUMP;

    //printf("Number of bytes: %d\n", number_of_bytes);

    printHexArray(buf + (*position), number_of_bytes, 0);
    printf("\n");

    return true;
}

bool checkExtensions(unsigned char* buf, long *position)
{
	int length = 0;
    int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		return false;
	}
   
    int start = *position;

    while ((*position) - start < length)
    {
        checkExtension(buf, position);
    }

    return true;
}

bool checkExtension(unsigned char* buf, long *position)
{
	int length = 0;
    int number_of_bytes = 0;

	if (!checkSequence(buf, position, &length, &number_of_bytes))
	{
		return false;
	}
    int dst = *position + length;
   
    char* objID = NULL;
	checkObjectIdentifier(buf, position, &objID);
	printf("\t\tExtension %s\n", objID);
	free(objID);

/*    bool critical = false;
    checkBoolean(buf, position, &critical);
	printf("\t\t\tCritical: %d\n", critical);

    char *str = NULL;
    checkOctetString(buf, position, str);
	printf("\t\t\tValue:");
    printfHexString(str);
    free(str)
*/

    printf("\t\t\tParameters: <not supported>\n");


    //printf("Number of bytes: %d\n", length);
    (*position) = dst;

    return true;
}

int main(int argc, char **argv)
{
	FILE *f_input;
	char *buf_name;
	
	if (argc <= 1) 
	{
		fprintf(stderr, "usage: %s <input file>\n", argv[0]);
		return 2;
	}
    buf_name = argv[1];
    f_input = fopen(buf_name, "rb");

	if(f_input == NULL)
	{
        perror("fopen");
		return 2;
	}
    //fprintf(stderr, "The file '%s' was opened\n", buf_name);

	
	fseek(f_input, 0, SEEK_END);
	fsize = ftell(f_input);
	fseek(f_input, 0, SEEK_SET);
	//fprintf(stderr, "Size of file: %ld Bytes.\n", fsize);

	unsigned char *buf = (unsigned char*) malloc(fsize + 1);
	ssize_t read = fread(buf, 1, fsize, f_input);
    if (read != fsize) {
        fclose(f_input);
        perror("fread");
		return 2;
    }

	buf[fsize] = '\0';

	fclose(f_input);
	//fprintf(stderr, "The file '%s' was closed\n", buf_name);

	
	long position = 0;
    int length, number_of_bytes = 0;
    if (!checkSequence(buf, &position, &length, &number_of_bytes) ||
        (length + number_of_bytes + 2 != fsize))
    {
        fail(buf);
    }
    //fprintf(stderr, "%d  %d  %ld\n", length, number_of_bytes, fsize);
    
    printf("TBS Certificate:\n");
    checkTBSCertificate(buf, &position);
    printf("Signature algorithm:\n");
    checkAlgorithmIdentifier(buf, &position);
    printf("Signature:\n");
	if (!checkBitString(buf, &position, &length, &number_of_bytes))
	{
        fail(buf);
	}

    printHexArray(buf + position, length, 1);
    position += length;
    printf("\n");

	free(buf);

    if (fsize != position)
    {
        fprintf(stderr, "Parsing failed\n");
        return 1;
    }
	return 0;
}
